﻿Imports Intergraph.GeoMedia
Imports Intergraph.GeoMedia.GDO
Imports Intergraph.GeoMedia.PBasic


Public Class CalculSolar
    Public objMV As Object
    Public objCS As PCSS.CoordSystem
    Public objDB As GDatabase
    Public objTbl As GTableDef
    Public objFld As GField
    Public objGSS As PClient.GeometryStorageService
    Public objOut As Object
    Public objLlumSolar As Object
    Public objRayTracing As RayTracing 'l'avaluació de l'ombra del propi edifici sobre els punts de la façana es fa amb Ray tracing sempre
    Public radi As Double
    Public nEdificis As Integer
    Public punts_per_costat As Int16
    Public HoresTotalsCoberta As Double
    Public HoresTotalsPisos As Double()
    Public mascaraCoberta As Double(), mascaraFasana As Double()
    Public algorisme As String
    Public objRSInput As GDO.GRecordset
    Public objRSInput2 As GDO.GRecordset
    Public objRSInput3 As GDO.GRecordset
    Public objRSOutput As GDO.GRecordset
    Dim objConnect As PClient.Connection
    Public IntervalTemps As Integer
    Private ProcesCancellat As Boolean = False

    'retorna diversos paràmetres per calcular la posició del sol o l'ortus i l'ocàs 
    Sub CalculaParametresSol(any As Long, mes As Long, dia As Long, hora As Long, minut As Long, longitud As Double, latitude As Double, ByRef diesAbans2000 As Double, ByRef lambda As Double, ByRef epsilon As Double, ByRef alpha As Double, ByRef delta As Double)
        'Dies abans del 2000
        diesAbans2000 = 367 * any - Int(7 * (any + Int((mes + 9) / 12)) / 4) + Int(275 * mes / 9) + dia + (hora + minut / 60) / 24 - 730531.5

        'longitud mitja del sol (L)
        Dim L As Double = (280.461 + 0.9856474 * diesAbans2000) Mod 360

        'Anomalia mitja del sol (g)
        Dim g As Double = (357.528 + 0.9856003 * diesAbans2000) Mod 360

        'Longitud eclíptica del sol (lambda)
        lambda = L + 1.915 * Math.Sin(g * Math.PI / 180) + 0.02 * Math.Sin(2 * g * Math.PI / 180)

        'Obliquitat del pla eclíptic (epsilon)
        epsilon = 23.439 - 0.0000004 * diesAbans2000

        'Ascensió recta (alpha)
        Dim Y As Double = Math.Cos(epsilon * Math.PI / 180) * Math.Sin(lambda * Math.PI / 180)
        Dim X As Double = Math.Cos(lambda * Math.PI / 180)

        alpha = Math.Atan2(Y, X) * 180 / Math.PI

        'Declinacio (delta)
        delta = Math.Asin(Math.Sin(epsilon * Math.PI / 180) * Math.Sin(lambda * Math.PI / 180))

    End Sub

    'calcula la posició del sol donades la data i l'hora
    Function CalculaAzimutAltitud(any As Long, mes As Long, dia As Long, hora As Long, minut As Long, longitud As Double, latitude As Double) As PBasic.point
        Dim diesAbans2000 As Double, lambda As Double, epsilon As Double, alpha As Double, delta As Double
        Dim pSol As PBasic.point = New PBasic.point

        'comprobant que l'hora d'entrada és entre l'ortus i l'ocàs
        Dim tuorto As Double, tuocas As Double
        CalculaOrtusOcas(any, mes, dia, longitud, latitude, tuorto, tuocas)
        If (hora + (minut + 10) / 60) < tuorto Or (hora + (minut - 10) / 60) > tuocas Then
            MsgBox("L'hora d'entrada no es troba entre l'ortus i l'ocàs! Operació cancel·lada.", MsgBoxStyle.Exclamation)
            pSol.X = 0
            pSol.Y = 0
            pSol.Z = 0
            Return psol
        End If

        CalculaParametresSol(any, mes, dia, hora, minut, longitud, latitude, diesAbans2000, lambda, epsilon, alpha, delta)

        'local sideral time (LST)
        Dim LST As Double = (280.46061837 + 360.98564736629 * diesAbans2000 + longitud) Mod 360

        Dim ha As Double = (LST - alpha) Mod 360

        'Altitud
        Dim altitud As Double = Math.Asin(Math.Sin(delta) * Math.Sin(latitude * Math.PI / 180) + Math.Cos(delta) * Math.Cos(latitude * Math.PI / 180) * Math.Cos(ha * Math.PI / 180))

        'Azimut (Z)
        Dim Y As Double = -Math.Cos(delta) * Math.Cos(latitude * Math.PI / 180) * Math.Sin(ha * Math.PI / 180)
        Dim X As Double = Math.Sin(delta) - Math.Sin(latitude * Math.PI / 180) * Math.Sin(altitud)
        Dim Z As Double = Math.Atan2(Y, X)

        'altitud = 20 * Math.PI / 180
        'Z = 90 * Math.PI / 180

        'calculem la posició del sol assumint que tenim un sistema de coordenades esfèriques centrades al punt d'interés
        Dim rSol As Double = 600000.0
        pSol.X = rSol * Math.Cos(altitud) * Math.Sin(Z)
        pSol.Y = rSol * Math.Cos(altitud) * Math.Cos(Z)
        pSol.Z = rSol * Math.Sin(altitud)

        Return pSol
    End Function

    'calcula les hores de l'ortus i l'ocàs
    Public Sub CalculaOrtusOcas(any As Long, mes As Long, dia As Long, longitud As Double, latitude As Double, ByRef TUORTO As Double, ByRef TUOCAS As Double)
        Dim diesAbans2000 As Double, lambda As Double, epsilon As Double, alpha As Double, delta As Double

        CalculaParametresSol(any, mes, dia, 12, 0, longitud, latitude, diesAbans2000, lambda, epsilon, alpha, delta)

        'Calculant ortus i ocàs
        Dim GGG As Double = 1
        If (any <= 1585) Then GGG = 0
        Dim JD = -1 * Math.Floor(7 * (Math.Floor((mes + 9) / 12) + any) / 4)
        Dim S As Double = 1
        If ((mes - 9) < 0) Then S = -1
        Dim A As Double = Math.Abs(mes - 9)
        Dim J1 As Double = Math.Floor(any + S * Math.Floor(A / 7))
        J1 = -1 * Math.Floor((Math.Floor(J1 / 100) + 1) * 3 / 4)
        JD = JD + Math.Floor(275 * mes / 9) + dia + (GGG * J1)
        JD = JD + 1721027 + 2 * GGG + 367 * any - 0.5
        Dim J2 As Double = JD

        Dim M0 As Double = 2.12344
        Dim MN As Double = 0.0172019
        Dim P As Double = 4.93204
        Dim VP As Double = 0.0000822
        Dim T0 As Double = 2444000.5
        P = P + (J2 - T0) * VP / 100
        Dim AM As Double = ((M0 + MN * (J2 - T0)) * 180 / Math.PI Mod 360) * Math.PI / 180
        Dim ET As Double = 0.016718
        Dim MR As Double = 0.04301
        Dim F As Double = 13750.987
        Dim C As Double = 2 * ET * F * Math.Sin(AM) + 1.25 * ET * ET * F * Math.Sin(2 * AM)
        Dim R As Double = -MR * F * Math.Sin(2 * (P + AM)) + MR * MR * F * Math.Sin(4 * (P + AM)) / 2
        ET = C + R

        ' calculant l'arc semi-diurni
        Dim h0 As Double = Math.Acos(-Math.Tan(latitude * Math.PI / 180) * Math.Tan(delta))
        'variació de la inclinació
        Dim VD As Double = 0.9856 * Math.Sin(epsilon * Math.PI / 180) * Math.Cos(lambda * Math.PI / 180) / Math.Cos(delta)
        'obtenint el temps de l'ortus
        Dim vdor As Double = VD * (-h0 * 180 / Math.PI + 180) / 360
        Dim DCOR As Double = delta + vdor * Math.PI / 180
        Dim HORTO As Double = -Math.Acos(-Math.Tan(latitude * Math.PI / 180) * Math.Tan(DCOR))
        Dim VHORTO As Double = 5 / (6 * Math.Cos(latitude * Math.PI / 180) * Math.Cos(DCOR) * Math.Sin(HORTO))
        HORTO = (HORTO * 180 / Math.PI + VHORTO) / 15
        TUORTO = HORTO + ET / 3600 - longitud / 15 + 12 + 10 / 60 'li afegim 10 minuts per tenir un marge i assegurar-nos que el sol es troba sobre l'horitzó

        'càlcul de l'ocàs
        Dim VDOC As Double = VD * (h0 * 180 / Math.PI + 180) / 360
        Dim DCOC As Double = delta * 180 / Math.PI + VDOC
        Dim HOC As Double = Math.Acos(-Math.Tan(latitude * Math.PI / 180) * Math.Tan(DCOC * Math.PI / 180))
        Dim VHOC As Double = 5 / (6 * Math.Cos(latitude * Math.PI / 180) * Math.Cos(DCOC * Math.PI / 180) * Math.Sin(HOC))
        HOC = (HOC * 180 / Math.PI + VHOC) / 15
        TUOCAS = HOC + ET / 3600 - longitud / 15 + 12 - 10 / 60 'li restem 10 minuts per tenir un marge i assegurar-nos que el sol es troba sobre l'horitzó

        'MsgBox("orto: " + Convert.ToString(Math.Floor(TUORTO)) + "h" + Convert.ToString((TUORTO - Math.Floor(TUORTO)) * 60 + 0.5) + "min" + _
        '       "ocàs: " + Convert.ToString(Math.Floor(TUOCAS)) + "h" + Convert.ToString((TUOCAS - Math.Floor(TUOCAS)) * 60 + 0.5) + "min")

    End Sub

    Public Overridable Function crearTaula() As Boolean
        'polimorfisme
        Return True
    End Function

    Public Overridable Sub CreacioPunts(objCoberta As PBasic.PolygonGeometry, objFasana As PBasic.PolygonGeometry, ByRef PuntsCoberta As PBasic.Points, ByRef PuntsFasana As PBasic.Points)
        'polimorfisme
    End Sub

    
    Public Overridable Sub AvaluaMascaraRaigSol()
        'polimorfisme
    End Sub

    'inicialitza les variables relacionades amb la base de dades
    Public Overridable Function InicialitzaTaules() As Boolean
        'Accedint a la base de dades
        objConnect = objGeoApp.Document.Connections(1)
        objDB = objConnect.Database

        Return True
    End Function

    'realitza el càlcul de la llum per a una hora donada
    Public Sub CalculaHora(PuntsCoberta As PBasic.Points, PuntsFasana As PBasic.Points, any As Long, mes As Long, dia As Long, hora As Long, minut As Long, longitud As Double, latitud As Double, pCentre As PBasic.point)
        Dim poligon As PBasic.PolygonGeometry = New PBasic.PolygonGeometry

        'calcula la posició del sol
        Dim pSol As PBasic.point = New PBasic.point
        pSol = CalculaAzimutAltitud(any, mes, dia, hora, minut, longitud, latitud)

        'Loop per cada edifici interferent
        objRSInput2.MoveFirst()
        Do While Not objRSInput2.EOF
            'mirem que no sigui l'edifici, del qual estem avaluant la llum
            If objRSInput2.GFields("ID_Edifici").Value <> objRSInput.GFields("ID_Edifici").Value Then
                'obtenint la geometria per treballar amb ella
                objGSS.StorageToGeometry(objRSInput2.GFields("Geometry").Value, poligon)
                'Avaluar si l'edifici és a dins d'un rang determinat. Ho fem només amb el primer polígon
                'Si és fora de rang, passem al següent edifici
                If objLlumSolar.EdificiValid(poligon, radi, pCentre, pSol) = True Then
                    'sel·leccionem els polígons
                    'com aquesta sel·lecció depèn de l'algorisme escollit, fem servir l'objecte perque torni la cerca
                    Dim comandaSelect As String = objLlumSolar.ComandaSelect(objRSInput2.GFields("ID_Edifici").Value)
                    objRSInput3 = objDB.OpenRecordset(comandaSelect)
                    'Bucle per cada polígon de l'edifici interferent
                    Do While Not objRSInput3.EOF
                        objGSS.StorageToGeometry(objRSInput3.GFields("Geometry").Value, poligon)

                        'Cacula les hores solar pel dia donat
                        objLlumSolar.AvaluaSolDirecte(PuntsCoberta, mascaraCoberta, poligon, pSol)
                        objLlumSolar.AvaluaSolDirecte(PuntsFasana, mascaraFasana, poligon, pSol)

                        objRSInput3.MoveNext()
                    Loop
                    objRSInput3.Close()
                End If
            End If

            objRSInput2.MoveNext()
            '.refresh()
        Loop

        'pels punts de la fasana avaluem també l'ombra del mateix edifici (coberta i les altres cares)
        'aquesta avaluació es fa sempre amb l'algorisme de ray tracing, ja que proporció de triangles no dóna els resultats esperats
        objRSInput3 = objDB.OpenRecordset("Select Geometry from Poligons3D where ID_Edifici = '" + objRSInput.GFields("ID_Edifici").Value + "' and Tipus_Poligon <> 'F'")
        'Loop per cada polígon
        Do While Not objRSInput3.EOF
            objGSS.StorageToGeometry(objRSInput3.GFields("Geometry").Value, poligon)
            objRayTracing.AvaluaSolDirecte(PuntsFasana, mascaraFasana, poligon, pSol)
            objRSInput3.MoveNext()
        Loop
        objRSInput3.Close()

        poligon = Nothing
    End Sub

    'inicialitza les màscares a 1
    Public Sub InicialitzaMascares()
        Dim j As Long
        'inicialitzant a 1 la màscara de la coberta
        For j = 0 To mascaraCoberta.Count - 1
            mascaraCoberta(j) = 1
        Next

        'inicialitzant a 1 la màscara de la fasana
        For j = 0 To mascaraFasana.Count - 1
            mascaraFasana(j) = 1
        Next
    End Sub

    'bucle per fer el càlcul de la llum solar per a un dia sencer
    Public Overridable Sub CalculaDia(PuntsCoberta As PBasic.Points, PuntsFasana As PBasic.Points, dia As Date, longitud As Double, latitud As Double, pCentre As PBasic.point)
        Dim TUORTO As Double, TUOCAS As Double

        'Calculant ortus i ocas
        CalculaOrtusOcas(dia.Year, dia.Month, dia.Day, longitud, latitud, TUORTO, TUOCAS)

        Dim HoraActual As DateTime = New DateTime(dia.Year, dia.Month, dia.Day, CInt(Math.Floor(TUORTO)), CInt((TUORTO - Math.Floor(TUORTO)) * 60), 0, 0)

        'número d'intervals dins del dia. IntervalTemps ve donat en minuts, mentre que tuorto i tuocas estan en hores
        Dim nIntervals As Long = Math.Floor((TUOCAS - TUORTO) / (IntervalTemps / 60))

        Dim HoresCoberta(nIntervals - 1) As Double
        Dim HoresPisos(nIntervals * PuntsFasana.Count - 1) As Double

        'Bucle per avaluar les hores totals de sol dins d'un dia amb l'interval d'hores donat
        For m As Long = 0 To nIntervals - 1
            'inicialitzant les màscares a 1
            InicialitzaMascares()

            'Cacula les hores solar pel dia donat
            CalculaHora(PuntsCoberta, PuntsFasana, HoraActual.Year, HoraActual.Month, HoraActual.Day, HoraActual.Hour, HoraActual.Minute, longitud, latitud, pCentre)

            'avaluem la màscara
            If Total(mascaraCoberta) / mascaraCoberta.Count * 100 > 25 Then HoresCoberta(m) = 1 'com a mínim el 25% dels punts a la coberta
            For n As Long = 0 To PuntsFasana.Count - 1
                HoresPisos(n * nIntervals + m) = mascaraFasana(n) 'només hi ha un punt per pis
            Next

            'afegint interval
            HoraActual = HoraActual.AddMinutes(IntervalTemps)
        Next

        HoresTotalsCoberta = HoresTotalsCoberta + Total(HoresCoberta) * IntervalTemps / 60
        For m As Long = 0 To PuntsFasana.Count - 1
            HoresTotalsPisos(m) = HoresTotalsPisos(m) + Total(HoresPisos.Skip(m * nIntervals).Take(nIntervals).ToArray()) * IntervalTemps / 60
        Next
    End Sub

    'funció per a controlar si l'usuari cancel·la el càlcul
    Private Sub Button1_Click(sender As Object, e As EventArgs)
        MsgBox("Procés cancel·lat per l'usuari!")
        ProcesCancellat = True
    End Sub

    'bucle per realitzar el càlcul de la lum per a un període donat de dies. 
    'És el bucle principal tant per a la funcionalitat de percentatge solar com per a la d'hores totals.
    Public Sub CalculaPeriode(DataInici As Date, DataFinal As Date)
        Dim objCoberta As PBasic.PolygonGeometry = New PBasic.PolygonGeometry
        Dim objFasana As PBasic.PolygonGeometry = New PBasic.PolygonGeometry
        Dim pRef As PBasic.point = New PBasic.point

        'Inicialitza les taules
        If InicialitzaTaules() = False Then Return

        'obrint recordsets
        objRSInput = objDB.OpenRecordset("Select * from Poligons3D where Tipus_Poligon = 'C'")
        objRSInput2 = objDB.OpenRecordset("Select * from Poligons3D where Tipus_Poligon = 'C'")

        'per saber el número de registres hem d'anar a l'últim primer!
        objRSInput.MoveLast()
        nEdificis = objRSInput.RecordCount
        objRSInput.MoveFirst()

        Dim progressBar As New MyProgressBar(nEdificis, "Calculant llum solar. ")
        AddHandler progressBar.Button1.Click, AddressOf Me.Button1_Click
        progressBar.Show()

        Dim m As Long = 0, n As Long
        For n = 1 To nEdificis
            'barra de progrés
            progressBar.Valor(n)

            'If objRSInput.GFields("ID_Edifici").Value <> "Edifici77" Then
            '    objRSInput.MoveNext()
            '    Continue For
            'End If

            'obtenint polígon corresponent a la coberta
            objGSS.StorageToGeometry(objRSInput.GFields("Geometry").Value, objCoberta)

            'obtenint el punt central de la coberta
            Dim pCentre As PBasic.point
            pCentre = PuntCentral(objCoberta)
            pRef.X = pCentre.X
            pRef.Y = pCentre.Y
            pRef.Z = pCentre.Z
            'obtenint longitud/latitud
            'Transforma les coordenades a latitud/longitud
            objCS.TransformPoint(PCSS.CSPointConstants.cspUOR, 1, PCSS.CSPointConstants.cspLLO, 1, pRef.X, pRef.Y, pRef.Z)
            Dim longitud As Double = pRef.X * 180 / Math.PI
            Dim latitud As Double = pRef.Y * 180 / Math.PI

            'seleccionem la façana de l'edifici actual
            objRSInput3 = objDB.OpenRecordset("Select Geometry from Poligons3D where ID_Edifici = '" + objRSInput.GFields("ID_Edifici").Value + "' and Tipus_Poligon = 'F'")
            objGSS.StorageToGeometry(objRSInput3.GFields("Geometry").Value, objFasana)
            objRSInput3.Close()

            Dim PuntsCoberta As PBasic.Points = New PBasic.Points
            Dim PuntsFasana As PBasic.Points = New PBasic.Points

            'Creem els punts 
            'per aquesta funció es fa servir polimorfisme, de manera que en funció de l'anàlisi (percentatge o hores totals) es cridarà a una funció o a una altra
            CreacioPunts(objCoberta, objFasana, PuntsCoberta, PuntsFasana)
            ReDim mascaraCoberta(PuntsCoberta.Count - 1) 'inicialització
            ReDim mascaraFasana(PuntsFasana.Count - 1) 'inicialització

            'inicialitzem el número total d'hores de llum (per funcionalitat número total d'hores)
            HoresTotalsCoberta = 0
            ReDim HoresTotalsPisos(PuntsFasana.Count - 1)

            'bucle per cada dia dins del període sel·leccionat
            Dim DiaActual As Date = DataInici
            While (DiaActual <= DataFinal)
                'càlcul pel dia donat
                CalculaDia(PuntsCoberta, PuntsFasana, DiaActual, longitud, latitud, pCentre)

                'següent dia
                DiaActual = DiaActual.AddDays(1)
            End While

            'de nou polimorfisme. Cada funcionalitat ho avalua de manera diferent. En aquesta funció també es crea el registre
            AvaluaMascaraRaigSol()

            objRSInput.MoveNext()

            PuntsCoberta = Nothing
            PuntsFasana = Nothing
            pCentre = Nothing

            'per a permetre que l'usuari cancel·li el càlcul
            System.Windows.Forms.Application.DoEvents()
            If ProcesCancellat = True Then Exit For
        Next

        objCoberta = Nothing
        objFasana = Nothing

        progressBar.Acabar()
        progressBar.Close()

        'tancant objectes i taules
        TancaObjectes()

        ''reobrint la connexió
        'objConnect.ReloadDatabase()
        'objConnect.BroadcastDatabaseChanges()
        'objGeoApp.ActiveWindow.MapView.Refresh(True)
        ''reobrint la connexió
        'objConnect.ReloadDatabase()
        'For Each LE As PView.LegendEntry In objGeoApp.ActiveWindow.MapView.Legend.LegendEntries
        '    LE.LoadData()
        'Next LE
        'objGeoApp.ActiveWindow.MapView.Legend.Refresh()
        'objGeoApp.ActiveWindow.MapView.Refresh()
    End Sub

    'tanca objectes
    Public Sub TancaObjectes()
        objRSInput.Close()
        objRSInput2.Close()
        objRSOutput.Close()
        objGSS = Nothing
        objRSInput = Nothing
        objRSInput2 = Nothing
        objRSInput3 = Nothing
        objRSOutput = Nothing
        objRayTracing = Nothing
        objLlumSolar = Nothing
    End Sub

    'constructor
    Public Sub New(algorisme_ As Long)
        objMV = objGeoApp.ActiveWindow.MapView
        objCS = objMV.CoordSystemsMgr.CoordSystem
        objGSS = New PClient.GeometryStorageService

        'creant objecte per fer el càlcul de la llum solar
        Select Case algorisme_
            Case 0
                objLlumSolar = New RayTracing
                algorisme = "Ray tracing"
            Case 1
                objLlumSolar = New ProporcioTriangles
                algorisme = "Proporció triangles"
            Case Else
                MsgBox("Algorisme no valid. Sel·leccionant Proporció Triangles.")
                objLlumSolar = New ProporcioTriangles
                algorisme = "Proporció triangles"
        End Select

        objRayTracing = New RayTracing
    End Sub
End Class
