﻿Imports Intergraph.GeoMedia
Imports Intergraph.GeoMedia.GDO
Imports Intergraph.GeoMedia.PBasic

Public Class HoresTotals
    Inherits CalculSolar
    Dim DataInici As Date
    Dim DataFinal As Date
    Dim nMaxPisos As Long   'màxim número de pisos

    'crea la taula HoresTotals si no existeix
    Overrides Function crearTaula() As Boolean

        If ExisteixTaula("HoresTotals", objDB) = True Then
            Dim objRSInput As GDO.GRecordset
            Dim valor As String = Convert.ToString(DataInici) + "-" + Convert.ToString(DataFinal)
            objRSInput = objDB.OpenRecordset("Select * from HoresTotals where Periode = '" + valor + "'")
            If objRSInput.EOF = False Then
                Dim Resultat As Long = MsgBox("La taula HoresTotals ja conté registres amb aquest periode. Si es procedeix s'esborraran tots els registres corresponents.", 49)
                If Resultat = 1 Then
                    EsborraRegistres(objRSInput)
                    Return True
                Else
                    'cancel·lat per l'usuari
                    objRSInput.Close()
                    Return False
                End If
            Else
                objRSInput.Close()
                'la taula existeix però no conté registres amb la data donada. Per tant, podem procedir
                Return True
            End If
        End If

        'Create table object 
        objTbl = objDB.CreateTableDef("HoresTotals")
        objTbl.Name = "HoresTotals"

        'Autonumèric
        objFld = objTbl.CreateField("ID1", GDO.GConstants.gdbLong)
        objFld.Attributes = GDO.GConstants.gdbAutoIncrField
        objTbl.GFields.Append(objFld)
        objFld = Nothing

        'Camp de text (edifici)
        objFld = objTbl.CreateField("ID_Edifici", GDO.GConstants.gdbText)
        With objFld
            .AllowZeroLength = False
            .Attributes = GDO.GConstants.gdbUpdatableField
            .Size = 80
            .Required = True
        End With
        objTbl.GFields.Append(objFld)
        objFld = Nothing

        'Periode de temps avaluat
        objFld = objTbl.CreateField("Periode", GDO.GConstants.gdbText)
        objTbl.GFields.Append(objFld)
        objFld = Nothing

        'Planta
        objFld = objTbl.CreateField("Planta", GDO.GConstants.gdbText)
        With objFld
            .Size = 1
            .Required = True
        End With
        objTbl.GFields.Append(objFld)
        objFld = Nothing


        'Geometria de la planta (punts avaluats)
        objFld = objTbl.CreateField("Geometry", GDO.GConstants.gdbSpatial)
        With objFld
            .Type = GDO.GConstants.gdbSpatial
            .SubType = GDO.GConstants.gdbAreal
            .Required = True
        End With
        objTbl.GFields.Append(objFld)
        objFld = Nothing

        'Hora
        objFld = objTbl.CreateField("HoresTotals", GDO.GConstants.gdbDouble)
        With objFld
            .Required = True
        End With
        objTbl.GFields.Append(objFld)
        objFld = Nothing

        'crear index amb increment i nom 
        Dim objIdx As GIndex, objIdxFld As GField

        objIdx = objTbl.CreateIndex("IDidx")

        objIdxFld = objIdx.CreateField("ID1")
        objIdx.GFields.Append(objIdxFld)
        objIdxFld = Nothing

        With objIdx
            .Name = "IDidx"
            .IgnoreNulls = False
            .Primary = True
            .Unique = True
        End With

        objTbl.GIndexes.Append(objIdx)
        objIdx = Nothing

        objDB.GTableDefs.Append(objTbl)

        Return True
    End Function

    'crea els punts sobre la coberta i els diferents pisos per a la funcionalitat d'hores totals de sol
    Overrides Sub CreacioPunts(objCoberta As PBasic.PolygonGeometry, objFasana As PBasic.PolygonGeometry, ByRef PuntsCoberta As PBasic.Points, ByRef PuntsFasana As PBasic.Points)
        'per la coberta agafem tots els punts disponibles
        Dim j As Long
        For j = 1 To objCoberta.Points.Count - 1
            PuntsCoberta.Add(objCoberta.Points.Item(j))
        Next

        'omplint pisos
        'Trobant els valors màxims i mínims de la fasana
        Dim xMax, xMin, yMax, yMin, zMax, zMin As Double
        MaxMin(ValorCoordenada(objFasana.Points, 0), xMax, xMin)
        MaxMin(ValorCoordenada(objFasana.Points, 1), yMax, yMin)
        MaxMin(ValorCoordenada(objFasana.Points, 2), zMax, zMin)

        Dim nPisos As Long = Math.Floor((zMax - zMin) / 3) 'assumim un pis són 3 metres
        If nPisos = 0 Then nPisos = 1 'com a mínim la planta baixa
        Dim p As PBasic.point = New PBasic.point
        p.X = (xMin + xMax) / 2
        p.Y = (yMin + yMax) / 2
        For j = 0 To nPisos - 1
            p.Z = zMin + j * 3
            PuntsFasana.Add(p)
        Next

        p = Nothing
     End Sub

    'funcions d'inicialització de les taules
    Public Overrides Function InicialitzaTaules() As Boolean
        'obrint les altres taules necessàries
        MyBase.InicialitzaTaules()

        'comprovem que la taula Poligons3D existeixi
        'creem la taula per guardar les ombres
        If crearTaula() = False Then Return False

        'obrint taula de sortida
        objRSOutput = objDB.OpenRecordset("HoresTotals", GDO.GConstants.gdbOpenDynaset)

        Return True
    End Function

    'crea el registre corresponent a l'edifici del cursor actual amb els valors calculats de les hores totals per a la coberta i cada pis
    Overrides Sub AvaluaMascaraRaigSol()
        'hores totals de sol per la coberta
        objRSOutput.AddNew()
        objRSOutput.GFields("ID_Edifici").Value = objRSInput.GFields("ID_Edifici").Value
        objRSOutput.GFields("Periode").Value = Convert.ToString(DataInici) + "-" + Convert.ToString(DataFinal)
        objRSOutput.GFields("Geometry").Value = objRSInput.GFields("Geometry").Value
        objRSOutput.GFields("Planta").Value = "C"
        objRSOutput.GFields("HoresTotals").Value = HoresTotalsCoberta
        objRSOutput.Update()

        'hores totals per cada pis
        For m As Long = 0 To mascaraFasana.Count - 1
            objRSOutput.AddNew()
            objRSOutput.GFields("ID_Edifici").Value = objRSInput.GFields("ID_Edifici").Value
            objRSOutput.GFields("Periode").Value = Convert.ToString(DataInici) + "-" + Convert.ToString(DataFinal)
            objRSOutput.GFields("Geometry").Value = objRSInput.GFields("Geometry").Value
            objRSOutput.GFields("Planta").Value = Convert.ToString(m)
            objRSOutput.GFields("HoresTotals").Value = HoresTotalsPisos(m)
            objRSOutput.Update()
        Next
    End Sub

    'funció principal per a fer el càlcul
    Public Sub Calcula()
        'comprobem que la taula Poligons3D existeixi a la base de dades
        If ExisteixTaula("Poligons3D", objGeoApp.Document.Connections(1).Database) = False Then
            MsgBox("La Taula Poligons3D no s'ha trobat a la base de dades. Has de crearla primer amb la utilitat 'Afegir alçades'. Operació cancel·lada.", MsgBoxStyle.Exclamation)
            Return
        End If

        'Cridem la funció principal, la qual només farà un únic dia
        CalculaPeriode(DataInici, DataFinal)
    End Sub

    'constructor
    Public Sub New(DataInici_ As Date, DataFinal_ As Date, TipusAlgorisme As String, radi_ As Double, IntervalTemps_ As Int64, algorisme_ As Long)
        MyBase.New(algorisme_)

        DataInici = DataInici_
        DataFinal = DataFinal_
        radi = radi_
        IntervalTemps = IntervalTemps_
        algorisme = TipusAlgorisme
    End Sub
End Class
