﻿Imports Intergraph.GeoMedia
Imports Intergraph.GeoMedia.GDO
Imports Intergraph.GeoMedia.PBasic

Public Class RayTracing

    Inherits LlumSolar

    'aquesta funció avalua si dos vectors donats són ortogonals
    Public Overrides Function AvaluaVectorsOrtogonals(poligon As PBasic.PolygonGeometry, pSol As PBasic.point, ByRef l As PBasic.Vector, ByRef n As PBasic.Vector) As Boolean
        'vector director del pla
        n = vectorDirector(poligon)

        'vector punt - sol
        'com el punt a avaluar és el de referència (el nostre centre de coordenades locals), aquest vector és la coordenada del sol amb signe invers
        l.I = -pSol.X
        l.J = -pSol.Y
        l.K = -pSol.Z
        l = objMath.Normalize(l)

        'mirar si el vector punt-sol és quasi ortogonal al vector director del pla. Si és així,
        'llavors no cal evaluar-lo per guanyar eficiència
        Return vectorsQuasiOrtogonals(l, n) 'torna true si ho són
    End Function

    'la funció avalua si un polígon dóna ombra o no al punt de referència fent servir l'algorisme de ray tracing
    Public Overrides Function AvaluaSolDirectePunt(punt As PBasic.point, pSol As PBasic.point, poligon As PBasic.PolygonGeometry, n As PBasic.Vector, l As PBasic.Vector, zMax As Double) As Integer

        'en primer lloc avaluem si el polígon és més baix que el punt que s'està avaluant
        'si ho és, no li pot fer ombra
        If zMax < punt.Z Then Return 1

        'Punt al polígon a avaluar. Agafem el primer
        Dim P0 As PBasic.point = New PBasic.point
        P0.X = poligon.Points.Item(1).X - punt.X
        P0.Y = poligon.Points.Item(1).Y - punt.Y
        P0.Z = poligon.Points.Item(1).Z - punt.Z

        'intersecció del vector punt-sol amb el pla del polígon
        Dim d As Double
        d = objMath.DotProduct(objMath.BuildVector(pSol, P0), n) / objMath.DotProduct(l, n)
        Dim pPlane As PBasic.point = New PBasic.point
        'intersecció vector pla definit per polígon. 
        pPlane.X = d * l.I + pSol.X
        pPlane.Y = d * l.J + pSol.Y
        pPlane.Z = d * l.K + pSol.Z

        'si l'alçada del punt on s'interessecciona és més petita que 0 (ja que aquesta Z és l'alçada respecte al 
        'punt de referencia) llavors no pot fer ombra, o el punt on intersecciona és més alt que l'edifici interferent
        If pPlane.Z < 0 Or (pPlane.Z > (zMax - punt.Z)) Then Return 1

        'la funció EdificiValid no sempre dóna els resultats esperats. 
        'per aquest motiu tornem a mirar la distàcia 2D (al pla X-Y) al sol. Si el punt és més lluny vol dir
        'que es troba darrera respecte al sol i per tant no dóna ombra.
        If distancia2D(pPlane, pSol) > Math.Sqrt(pSol.X ^ 2 + pSol.Y ^ 2) Then Return 1

        'afegim les coordenades del punt, ja que és el nostre centre de coordenades
        pPlane.X = pPlane.X + punt.X
        pPlane.Y = pPlane.Y + punt.Y
        pPlane.Z = pPlane.Z + punt.Z

        'mirant si el punt es troba dins el polígon
        Dim dins As Long
        dins = objMath.PointInPolygon(pPlane, poligon.Points)

        l = Nothing
        n = Nothing
        pPlane = Nothing
        P0 = Nothing
        'si no és a dins, vol dir que el raig no dóna a aquest polígon i per tant el punt reb la llum del sol
        If dins = -1 Then Return 1 Else Return 0
    End Function

End Class
