﻿Imports Intergraph.GeoMedia.GDO

Module Utilitats

    'retorna True si els vectors són gairebé ortogonals (angle entre 85º i 105º)
    Public Function vectorsQuasiOrtogonals(p1, p2) As Boolean
        'no cal dividir entre les magnituds ja que tots dos vectors ja han estat normalitzats
        Dim alpha As Double = Math.Abs(Math.Acos(objMath.DotProduct(p1, p2))) * 180 / Math.PI

        'si l'angle entre tots dos es troba entre 85 i 105 graus, el polígon és gairebé paral·lel al raig de sol
        'i per tant no produirà ombra sobre l'edifici d'interés
        If alpha > 85 And alpha < 105 Then Return True

        Return False
    End Function

    'retorna la distància euclidiana entre dos punts
    Public Function distancia(punt1 As PBasic.point, punt2 As PBasic.point) As Double
        Dim d As Double = Math.Sqrt((punt1.X - punt2.X) ^ 2 + (punt1.Y - punt2.Y) ^ 2 + (punt1.Z - punt2.Z) ^ 2)
        Return d
    End Function

    'retorna la distància euclidiana al pla X-Y entre dos punts
    Public Function distancia2D(punt1 As PBasic.point, punt2 As PBasic.point) As Double
        Dim d As Double = Math.Sqrt((punt1.X - punt2.X) ^ 2 + (punt1.Y - punt2.Y) ^ 2)
        Return d
    End Function

    'esborra tots els registres del cursor
    Public Sub EsborraRegistres(objRSInput As GDO.GRecordset)
        'esborra els registres corresponents
        objRSInput.MoveFirst()

        Do While Not objRSInput.EOF
            objRSInput.Delete()
            objRSInput.MoveNext()
        Loop

        objRSInput.Close()
    End Sub

    'Esborra tots els registres dins la taula donada
    Public Sub EsborraTotsElsRegistres(nom As String, objDB As GDatabase)
        Dim objRSInput As GDO.GRecordset

        'esborrar tots els registres
        objRSInput = objDB.OpenRecordset(nom, GDO.GConstants.gdbOpenDynaset)

        Do While Not objRSInput.EOF
            objRSInput.Delete()
            objRSInput.MoveNext()
        Loop

        objRSInput.Close()
    End Sub

    'avalua si la taula ja existeix a la base de dades
    Public Function ExisteixTaula(nom As String, objDB As GDatabase) As Boolean
        On Error GoTo ErrorHandler
        Dim Existeix As String = objDB.GTableDefs(nom).Name

        Return True

ErrorHandler:  'no afegim l'edifici
        If Err.Number <> 0 Then
            Return False
        End If

    End Function

    'retorna la suma de tots els components de l'array
    Public Function Total(v As Double()) As Double
        'retorna la suma dels valors dins l'array
        Dim m As Long
        Dim out As Double = 0
        For m = 0 To v.Count - 1
            out = out + v(m)
        Next
        Return out
    End Function

    'retorna un array amb els valors de la coordenada seleccionada
    Public Function ValorCoordenada(Points As PBasic.Points, coordinate As Integer) As Double()
        Dim values(Points.Count - 1) As Double
        Dim m As Long
        For m = 1 To Points.Count
            Select Case coordinate
                Case 0
                    values(m - 1) = Points.Item(m).X
                Case 1
                    values(m - 1) = Points.Item(m).Y
                Case 2
                    values(m - 1) = Points.Item(m).Z
            End Select

        Next
        Return values
    End Function

    'retorna els valors màxim i mínim dins d'un array
    Public Sub MaxMin(values As Double(), ByRef max As Double, ByRef min As Double)
        Dim m As Long

        For m = 0 To values.Count - 1
            If m = 0 Then
                max = values(m)
                min = values(m)
            Else
                If values(m) > max Then max = values(m)
                If values(m) < min Then min = values(m)
            End If
        Next
    End Sub

    'afegeix un punt al polígon
    Public Sub afegirPunt(objEdifici As PBasic.PolygonGeometry, x As Double, y As Double, z As Double)
        Dim objPoint As PBasic.point

        objPoint = New PBasic.point
        objPoint.X = x
        objPoint.Y = y
        objPoint.Z = z
        objEdifici.Points.Add(objPoint)
        objPoint = Nothing
    End Sub

    'calcula el punt geomètric central d'un polígon
    Public Function PuntCentral(poligon As PBasic.PolygonGeometry) As PBasic.point
        Dim xMax, xMin, yMax, yMin, zMax, zMin As Double
        Return PuntCentral(poligon, xMin, xMax, yMin, yMax, zMin, zMax)
    End Function

    'calcula el punt geomètric central d'un polígon
    Public Function PuntCentral(poligon As PBasic.PolygonGeometry, ByRef xMin As Double, ByRef xMax As Double, ByRef yMin As Double, ByRef yMax As Double, ByRef zMin As Double, ByRef zMax As Double) As PBasic.point
        MaxMin(ValorCoordenada(poligon.Points, 0), xMax, xMin)
        MaxMin(ValorCoordenada(poligon.Points, 1), yMax, yMin)
        MaxMin(ValorCoordenada(poligon.Points, 2), zMax, zMin)

        Dim p As PBasic.point = New PBasic.point

        'el punt central és el valor mig entre el màxim i el mínim de cada coodenada
        p.X = (xMin + xMax) * 0.5
        p.Y = (yMin + yMax) * 0.5
        p.Z = (zMin + zMax) * 0.5

        Return p
    End Function

    'retorna el vector director d'un polígon
    Public Function vectorDirector(poligon As PBasic.PolygonGeometry) As PBasic.Vector
        'agafem els tres primers punts i per calcular dos vectors
        Dim v1 As PBasic.Vector = New PBasic.Vector
        Dim v2 As PBasic.Vector = New PBasic.Vector

        'generem els dos vector
        v1 = objMath.BuildVector(poligon.Points.Item(2), poligon.Points.Item(1))
        v2 = objMath.BuildVector(poligon.Points.Item(poligon.Points.Count - 1), poligon.Points.Item(1))

        'el producte vectorial ens dóna el vector director del polígon
        Dim v3 As PBasic.Vector = New PBasic.Vector
        v3 = objMath.CrossProduct(v1, v2)
        'normalitzant
        'v3 = objMath.Normalize(v3)
        Dim modul As Double = Math.Sqrt(v3.I ^ 2 + v3.J ^ 2 + v3.K ^ 2)
        v3.I = v3.I / modul
        v3.J = v3.J / modul
        v3.K = v3.K / modul
        Return v3
    End Function
End Module
